package com.mcsnowflake.worktimer.notifications

import android.app.Notification
import android.app.NotificationChannel
import android.app.NotificationManager
import android.app.PendingIntent
import android.content.ContentResolver
import android.content.Context
import android.content.Intent
import android.media.AudioAttributes
import android.net.Uri
import androidx.core.app.NotificationCompat
import androidx.core.app.NotificationManagerCompat
import androidx.datastore.core.DataStore
import androidx.datastore.preferences.core.Preferences
import com.mcsnowflake.worktimer.*
import com.mcsnowflake.worktimer.MainActivity
import com.mcsnowflake.worktimer.R
import com.mcsnowflake.worktimer.TextHandler.getFormattedTime
import com.mcsnowflake.worktimer.TextHandler.getInfoText
import com.mcsnowflake.worktimer.TextHandler.getTitle
import com.mcsnowflake.worktimer.misc.getOrDefault
import com.mcsnowflake.worktimer.service.TimerAction
import com.mcsnowflake.worktimer.service.TimerService
import com.mcsnowflake.worktimer.service.TimerState.Running
import com.mcsnowflake.worktimer.service.TimerState.Running.*
import com.mcsnowflake.worktimer.service.WorkTimer.SessionType.*
import java.time.Duration
import kotlin.math.abs
import kotlinx.coroutines.runBlocking

class NotificationHandler(
    private val applicationContext: Context,
    private val notificationManager: NotificationManager,
    private val preferences: DataStore<Preferences>
) {

    private val pendingIntent = PendingIntent.getActivity(
        applicationContext,
        PendingRequestCode,
        Intent(applicationContext, MainActivity::class.java).addCategory(Intent.CATEGORY_LAUNCHER).setAction("android.intent.action.MAIN"),
        PendingIntent.FLAG_IMMUTABLE
    )

    fun getStartNotification(): Notification = NotificationCompat.Builder(applicationContext, AlarmChannelID)
        .setContentIntent(pendingIntent)
        .setSilent(true)
        .setSmallIcon(R.drawable.ic_logo_simple)
        .setContentTitle("WorkTimer is starting")
        .build()

    fun process(event: Running) {
        when (event) {
            is Started, is Created -> {
            }
            is Progress -> {
                if (runBlocking { preferences.getOrDefault(HYDRATION_MODE) } && event.isTimeToRemind())
                    sendHydrationReminder()
                sendProgressUpdate(event)
            }
            is Finished -> sendAlarmNotification(event)
            is Terminated -> sendStoppedNotification()
        }
    }

    private fun sendHydrationReminder() = notificationManager.notify(HydrationNotificationID, buildHydrateNotification())

    private fun sendProgressUpdate(update: Progress) = notificationManager.notify(TimerNotificationID, buildProgressNotification(update))

    private fun sendAlarmNotification(event: Finished) = notificationManager.notify(TimerNotificationID, buildAlarmNotification(event))

    private fun sendStoppedNotification() = notificationManager.notify(TimerNotificationID, buildStoppedNotification())

    // notification builder

    private fun buildProgressNotification(update: Progress): Notification {
        val builder = NotificationCompat.Builder(applicationContext, AlarmChannelID)
            .setContentIntent(pendingIntent)
            .setSilent(true)
            .setSmallIcon(R.drawable.ic_logo_simple)
            .addAction(buildNotificationAction(TimerAction.STOP))
            .setCategory(Notification.CATEGORY_PROGRESS)
            .setContentTitle(getTitle(update))
            .setContentText("${getFormattedTime(update)} remaining")
            .setProgress(100, update.progress.times(100).toInt(), false)

        return when (update.sessionType) {
            WORK_SESSION -> builder.addAction(buildNotificationAction(TimerAction.PAUSE))
            SHORT_BREAK, LONG_BREAK -> builder.addAction(buildNotificationAction(TimerAction.WORK))
        }.build()
    }

    private fun buildAlarmNotification(update: Running): Notification {
        val builder = NotificationCompat.Builder(applicationContext, AlarmChannelID)
            .setContentIntent(pendingIntent)
            .setSmallIcon(R.drawable.ic_logo_simple)
            .addAction(buildNotificationAction(TimerAction.STOP))
            .addAction(buildNotificationAction(TimerAction.RESUME))
            .setAllowSystemGeneratedContextualActions(false)
            .setCategory(Notification.CATEGORY_ALARM)
            .setContentTitle(getTitle(update))
            .setContentText(getInfoText(update))

        return if (runBlocking { preferences.getOrDefault(OVERTIME_MODE) })
            builder.addAction(buildNotificationAction(TimerAction.MORE)).build()
        else builder.build()
    }

    private fun buildStoppedNotification() = NotificationCompat.Builder(applicationContext, AlarmChannelID)
        .setContentIntent(pendingIntent)
        .setTimeoutAfter(100)
        .setAutoCancel(true)
        .setSmallIcon(R.drawable.ic_logo_simple)
        .setCategory(Notification.CATEGORY_ALARM)
        .setContentTitle("Timer stopped")
        .build()

    private fun buildHydrateNotification() = NotificationCompat.Builder(applicationContext, HydrateChannelID)
        .setAutoCancel(true)
        .setTimeoutAfter(10_000)
        .setSmallIcon(R.drawable.ic_logo_simple)
        .setCategory(Notification.CATEGORY_ALARM)
        .setContentTitle("Remember to hydrate yourself") // TODO refactor out string representations
        .build()

    private fun buildNotificationAction(action: TimerAction): NotificationCompat.Action = NotificationCompat.Action.Builder(
        action.getActionIcon(),
        action.getActionTitle(),
        PendingIntent.getService(applicationContext, PendingRequestCode, TimerService.getIntent(applicationContext, action.serviceCommand), PendingIntent.FLAG_IMMUTABLE)
    ).build()

    companion object {
        const val PendingRequestCode = 52334
        const val TimerNotificationID = 2345
        const val HydrationNotificationID = 5346
        const val AlarmChannelID = "47657"
        const val HydrateChannelID = "5253432"

        fun createChannelsWith(applicationContext: Context) {
            NotificationManagerCompat.from(applicationContext)
                .createNotificationChannel(
                    NotificationChannel(
                        AlarmChannelID,
                        "WorkTimer Notification Channel",
                        NotificationManager.IMPORTANCE_HIGH
                    ).apply {
                        lockscreenVisibility = NotificationCompat.VISIBILITY_PUBLIC
                        setShowBadge(false)
                    }
                )
            NotificationManagerCompat.from(applicationContext)
                .createNotificationChannel(
                    NotificationChannel(
                        HydrateChannelID,
                        "Hydration Reminder Channel",
                        NotificationManager.IMPORTANCE_HIGH
                    ).apply {
                        lockscreenVisibility = NotificationCompat.VISIBILITY_PUBLIC
                        setShowBadge(false)
                        setSound(
                            Uri.parse(ContentResolver.SCHEME_ANDROID_RESOURCE + "://" + applicationContext.packageName + "/" + R.raw.drop),
                            AudioAttributes.Builder()
                                .setContentType(AudioAttributes.CONTENT_TYPE_SONIFICATION)
                                .setUsage(AudioAttributes.USAGE_NOTIFICATION_RINGTONE)
                                .build()
                        )
                    }
                )
        }
    }

    private fun Progress.isTimeToRemind() =
        (sessionType == WORK_SESSION && abs(progress - 50) < 40 && remainingTime.toMinutes().rem(15) == 0L) ||
            (sessionType != WORK_SESSION && remainingTime == Duration.ofMinutes(2))
}
